"""Import local modules"""

import logging
import sys

from . import losses
from .losses import (
    add_to_metrics,
    build_loss_function,
    describe_metric,
    get_all_metrics,
    ALL_METRICS as metrics
)
from . import optim
from .optim import (
    assign_trunc_nums,
    CircuitSampler,
    run_BFGS,
    run_optimization,
    run_SGD,
    trunc_num_heuristic
)

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    __version__ = "unknown"
    version_tuple = (0, 0, "unknown")


qd_logger = logging.getLogger(__name__)


def get_logger() -> logging.Logger:
    """Get the qubit_discovery-wide parent logger."""
    return qd_logger


def log_to_stdout(level: int =logging.INFO) -> logging.Logger:
    """Set the qubit_discovery package to log to stdout.

    Parameters
    ----------
        level:
            The minimum level of logs to log.

    Returns
    ----------
        The logger.
    """
    handler = logging.StreamHandler(stream=sys.stdout)
    formatter = logging.Formatter('[%(asctime)s] %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')

    handler.setFormatter(formatter)
    qd_logger.addHandler(handler)
    qd_logger.setLevel(level)

    return qd_logger



"""
Get a summary of the circuit metrics and loss valuations from saved circuit
files after optimizations, and output as a text file.

Usage:
  circuit_summary.py <yaml_file> [--ids=<ids>] [--circuit_code=<circuit_code>]
                     [--optim_type=<optim_type>]
  circuit_summary.py  -h | --help
  circuit_summary.py  --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.

Options:
  -h, --help    Show this screen.
  --version     Show version.

  -i, --ids=<ids>                           Comma-delimited list of circuits to
                                            summarize.
  -c, --circuit_code=<circuit_code>         Circuit code.
  -o, --optim_type=<optim_type>             Optimization method used.
"""

from docopt import docopt
import qubit_discovery as qd
from qubit_discovery.losses import calculate_loss_metrics
import SQcircuit as sq

import analysis as an
from inout import (
    add_command_line_keys,
    Directory,
    load_yaml_file,
    get_metrics_dict,
    get_units
)
from utils import load_final_circuit

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
SUMMARY_REQUIRED_KEYS = [
    'ids',
    'circuit_code',
    'optim_type',
]

################################################################################
# Main.
################################################################################


def main():

    ############################################################################
    # Loading the Yaml file and command line parameters.
    ############################################################################

    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=SUMMARY_REQUIRED_KEYS,
    )

    directory = Directory(parameters, arguments['<yaml_file>'])

    ############################################################################
    # Summarize the Circuit.
    ############################################################################

    sq.set_engine('PyTorch')
    qd.log_to_stdout()

    for id_num in parameters['ids'].split(','):
        cr = load_final_circuit(directory.get_record_file_dir(
                record_type='circuit',
                circuit_code=parameters['circuit_code'],
                idx=id_num,
        ))
        cr.update()  # rebuild op memory
        cr.diag(parameters['num_eigenvalues'])

        metrics_in_optim, metrics_not_in_optim = get_metrics_dict(parameters)
        # add in element sensitivity for summary file
        # metrics_not_in_optim.append('element_sensitivity')

        # Prepare summary text for the circuit.
        summary_text = f'Description:\n{cr.description(_test=True)}\n'
        summary_text += an.build_circuit_topology_string(cr)

        summary_text += f'\ntrunc_nums: {cr.trunc_nums}\n'

        total_loss, loss_details, metrics = calculate_loss_metrics(
            circuit=cr,
            use_losses=parameters['use_losses'],
            use_metrics=metrics_not_in_optim
        )
        ########################################################################
        # In optimization losses
        ########################################################################
        loss_in_optim_summary = 'Loss in Optimization:\n'
        for key in metrics_in_optim:
            if loss_details[key + '_loss'] == 0.0:
                continue
            loss_in_optim_summary += (
                f"{key + '_loss'}: {loss_details[key + '_loss']}\n"
            )
        loss_in_optim_summary += (
            f'{"total_loss"}: {loss_details["total_loss"]}\n'
        )
        ########################################################################
        # Not in optimization losses
        ########################################################################
        loss_not_in_optim_summary = 'Other Losses:\n'
        for key in metrics_not_in_optim:
            if loss_details[key + '_loss'] == 0.0:
                continue
            loss_not_in_optim_summary += (
                f'{key + "_loss"}: {loss_details[key+ "_loss"]}\n'
            )
        ########################################################################
        # All the metrics
        ########################################################################
        metrics_summary = '\nMetrics:\n'
        for key in metrics_in_optim + metrics_not_in_optim:
            metrics_summary += f"{key} {get_units()[key]}: {metrics[key]}\n"

        summary_text += (
            metrics_summary +
            '\n' +
            loss_in_optim_summary +
            '\n' +
            loss_not_in_optim_summary
        )

        with open(
            directory.get_summary_file_dir(parameters['circuit_code'], id_num),
            'w'
        ) as f:
            f.write(summary_text)


if __name__ == '__main__':
    main()



"""Test Module for BFGS algorithm. CURRENTLY DEPRECATED"""

import numpy as np
import pytest
import SQcircuit as sq
import torch

from qubit_discovery import (
    build_loss_function,
    run_BFGS
)
from qubit_discovery.tests.conftest import (
    get_fluxonium,
    get_bounds,
    are_loss_dicts_close
)

@pytest.mark.skip('Currently deprecated')
def test_bfgs_run() -> None:
    """Test two steps of the BFGS algorithm."""

    target_params = [1.7817e-14, 3.4925e-07, 6.2715e+09]

    target_loss_record = {
        'frequency_loss': [np.array(1.e-13), np.array(1.e-13), np.array(1.e-13)],
        'number_of_gates_loss': [np.array(2.25816574e-05), np.array(2.2558016e-05), np.array(1.19195123e-05)],
        'flux_sensitivity_loss': [np.array(1.e-13), np.array(1.e-13), np.array(1.e-13)],
        'charge_sensitivity_loss': [np.array(1.e-13), np.array(1.e-13), np.array(1.e-13)],
        'total_loss': [np.array(2.25816577e-05), np.array(2.25580163e-05), np.array(1.19195126e-05)]
    }

    my_loss_function = build_loss_function(
        use_losses={
                "frequency": 1.0,
                "number_of_gates": 1.0,
                "flux_sensitivity": 1.0,
                "charge_sensitivity": 1.0,
            },
        use_metrics=[]
    )

    sq.set_engine('PyTorch')

    total_trunc_num = 120

    circuit = get_fluxonium()
    circuit.set_trunc_nums([total_trunc_num])

    circuit, loss_record, _ = run_BFGS(
        circuit=circuit,
        loss_metric_function=my_loss_function,
        max_iter=2,
        total_trunc_num=total_trunc_num,
        bounds = get_bounds(),
        num_eigenvalues=10,
        lr=1.0,
        tolerance=1e-7
    )

    print(circuit.parameters)
    print('Final', loss_record)
    assert torch.stack(circuit.parameters).detach() == pytest.approx(target_params, rel=1e-2)
    assert are_loss_dicts_close(loss_record, target_loss_record, rel=1e-2)
